//
// cell.c
//

#include <stdlib.h>
#include <math.h>

#include "cgmin.h"

extern g_num;

static int n;
static double size;
static long *list;
static long ***cell;
static int xxn, yyn, zzn;
static long cp;
static int status;
static long index[14];
static bool out;

long p1, p2;

void cellAlloc( int hn, double s )
{
  int i,j;

  n = (hn<=3)?6:2*hn;
  size = s;

  list = (long *)malloc( g_num * sizeof(long) );

  cell = (long ***)malloc( n * sizeof(long **) );
  for( i=0; i<n; i++ )
  {
    cell[i] = (long **)malloc( n * sizeof(long *) );
    for( j=0; j<n; j++ )
    {
      cell[i][j] = (long *)malloc( n * sizeof(long) );
    }
  }
}

void cellFree()
{
  int i,j;

  free( list );

  for( i=0; i<n; i++ )
  {
    for( j=0; j<n; j++ )  free( cell[i][j] );
    free( cell[i] );
  }
  free( cell );
}

void cellUpdate(double *y)
{
  long i,j,k;

  //
  // clear arrays
  //
  for( i=0; i<g_num; i++ ) list[i] = -1;
  for( i=0; i<n; i++ )
    for( j=0; j<n; j++ )
      for( k=0; k<n; k++ ) cell[i][j][k] = -1;

  //
  // update arrays
  //
  for( i=0; i<g_num; i++ )
  {
    int nx,ny,nz;
    long d;

    nx = (int)floor( y[i*3] / size ) + n/2;
    ny = (int)floor( y[i*3+1] / size ) + n/2;
    nz = (int)floor( y[i*3+2] / size ) + n/2;

    if( nx > n-2 || ny > n-2 || nz > n-2 
	|| nx < 1 || ny < 1 || nz < 1 )
    {
      printf("\nParticle out of boundary.\n");
      printf("Please increase the box size.\n\n");
      exit(1);
    }

    d = cell[ nx ][ ny ][ nz ];

    if( d == -1 )   // this is the first particle in this cell
    {
      cell[ nx ][ ny ][ nz ] = i;
    }
    else
    {
      while( list[ d ]!= -1 )      // search till the tail
      {
	d = list[ d ];
      }

      list[ d ] = i;
    }
  }

  cellReset();
}

void cellInit( bool genCp )
{
  if( genCp ) cp = cell[ xxn ][ yyn ][ zzn ];
  if( cp == -1 ) return;

  status = 0;
  index[ 0 ] = list[ cp ];      // next particle in current cell
  // head of 13 neighbors
  index[ 1 ] = cell[ xxn+1 ][ yyn-1 ][ zzn-1 ];
  index[ 2 ] = cell[ xxn+1 ][ yyn ][ zzn-1 ];
  index[ 3 ] = cell[ xxn+1 ][ yyn+1 ][ zzn-1 ];
  index[ 4 ] = cell[ xxn ][ yyn+1 ][ zzn-1 ];

  index[ 5 ] = cell[ xxn+1 ][ yyn-1 ][ zzn ];
  index[ 6 ] = cell[ xxn+1 ][ yyn ][ zzn ];
  index[ 7 ] = cell[ xxn+1 ][ yyn+1 ][ zzn ];
  index[ 8 ] = cell[ xxn ][ yyn+1 ][ zzn ];

  index[ 9 ] = cell[ xxn+1 ][ yyn-1 ][ zzn+1 ];
  index[ 10 ] = cell[ xxn+1 ][ yyn ][ zzn+1 ];
  index[ 11 ] = cell[ xxn+1 ][ yyn+1 ][ zzn+1 ];
  index[ 12 ] = cell[ xxn ][ yyn+1 ][ zzn+1 ];

  index[ 13 ] = cell[ xxn ][ yyn ][ zzn+1 ];
}

void cellReset()
{
  //
  // prepare for pair generating
  //
  out = false;
  xxn = 1;   yyn = 1;   zzn = 1;
  cellInit( true );
}

bool cellCheck()
{
  if( out ) return false;

  //
  // Step to next cell if no more particles in current cell
  //
  if( cp == -1 )
  {
    if( xxn < n-2 ) xxn++;  // skip the last one for programming simplicity
    else
    {
      if( yyn < n-2 ) { yyn++; xxn = 1; }
      else
      {
	if( zzn < n-2 ) { zzn++; yyn = 1; xxn = 1; }
	else { out = true; return false; }    // no more pairs
      }
    }

    cellInit( true );
    return cellCheck();
  } 

  //
  // Step to next status
  //
  if( index[ status ] == -1 )    // cp > -1
  {
    if( status < 13 ) status ++;  // step to next status
    else  // status == 13
    {
      cp = list[ cp ];
      cellInit( false ); 
    }

    return cellCheck();
  }

  return true;
}

bool cellGen()
{
  if( !cellCheck() ) return false;

  //
  // Generate a particle pair
  //
  p1 = cp;   p2 = index[ status ];
  index[ status ] = list[ p2 ];

  return true;
}
